/*
 * Robot Pidzy
 *
 * Christophe Le Lann <totofwebcreation@gmail.com>
 * http://www.totofweb.net/robots-projet-55.html
 *
 * Dimensions :
 *  - Rayon roue = 30mm
 *  - Rayon robot = 80mm
 *
 * Conversions :
 *  - 1 pas = pi/32 mm
 *  - 1 pas/interruption = 5pi/8 mm/s
 *  - 90 = 80pi/2 mm = 1280 pas
 *
 * Cette version effectue une srie de test de valeurs de
 * Kp, Ki et Kd d'un asservissement PID de vitesse sur un
 * seul moteur.
 * Le but est d'en sortir une srie de courbes typiques et
 * de vrifier les rsultats de la mthode de Ziegler Nichols
 *
 * Ceci n'est qu'un programme de test et ne suffit pas 
 * raliser un asservissement total d'un robot sur consigne
 * d'une carte matre (il faut modifier un peu le code pour a).
 */

//******************************
// Defines
//******************************

#define F_CPU 16000000UL

//******************************
// Attribution des IO
//******************************

// Encodeurs
#define ENC_G_COUNT_DDR		DDRC
#define ENC_G_COUNT_PORTIN	PINC
#define ENC_G_COUNT_PIN		6
#define ENC_G_SENS_DDR		DDRD
#define ENC_G_SENS_PORTIN	PIND
#define ENC_G_SENS_PIN		2
#define ENC_D_COUNT_DDR		DDRC
#define ENC_D_COUNT_PORTIN	PINC
#define ENC_D_COUNT_PIN		7
#define ENC_D_SENS_DDR		DDRD
#define ENC_D_SENS_PORTIN	PIND
#define ENC_D_SENS_PIN		3

// Moteurs
#define MOT_G_IN1_DDR		DDRC
#define MOT_G_IN1_PORT		PORTC
#define MOT_G_IN1_PIN		5
#define MOT_G_IN2_DDR		DDRC
#define MOT_G_IN2_PORT		PORTC
#define MOT_G_IN2_PIN		4
#define MOT_G_PWM_DDR		DDRD
#define MOT_G_PWM_PORT		PORTD
#define MOT_G_PWM_PIN		4
#define MOT_D_IN1_DDR		DDRC
#define MOT_D_IN1_PORT		PORTC
#define MOT_D_IN1_PIN		3
#define MOT_D_IN2_DDR		DDRC
#define MOT_D_IN2_PORT		PORTC
#define MOT_D_IN2_PIN		2
#define MOT_D_PWM_DDR		DDRD
#define MOT_D_PWM_PORT		PORTD
#define MOT_D_PWM_PIN		5

// BarGraph
#define BARG_RST_DDR		DDRA
#define BARG_RST_PORT		PORTA
#define BARG_RST_PIN		2
#define BARG_CNT_DDR		DDRA
#define BARG_CNT_PORT		PORTA
#define BARG_CNT_PIN		1

// Buffer srie
#define BUFFER_SIZE			128

// Jauge d'tat de la batterie
#define JAUGE_CAN_CHANNEL	0
#define JAUGE_MAXI			149		// Maxi 8.0V, 2^10*U*R/5, avec R=1/11
#define JAUGE_MINI			127		// Mini 6.8V, 2^10*U*R/5, avec R=1/11

//******************************
// Librairies
//******************************

#include <avr/io.h>
#include <avr/interrupt.h>
#include "bitmanip.c"
#include "encodeurs.c"
#include "bargraph.c"
#include "buffer.c"
#include "delais.c"
#include "usart.c"

//******************************
// Variables globales
//******************************

// Jauge d'tat de la batterie
unsigned char Jauge_overflow;		// Compteur divisant par 256 la frquence de mesure
float Kp, Ki, Kd;					// Coefficients de correction
signed long V_Consigne;				// Consigne  atteindre
signed long P_Consigne;				// Consigne  atteindre
unsigned char recalculer = 0;		// Passe  1 lorsqu'il faut recalculer la correction
unsigned int Duree;					// Duree de session d'enregistrement (en nombre de 50ms)
unsigned char count;				// Variable utilise pour compter les interruptions de 10ms


// PID
typedef struct struct_PID {
	signed long position_old;
	signed long vitesse;
	signed long erreur_vitesse;
	signed long erreur_vitesse_old;
	signed long consigne_vitesse;
	signed long integral_vitesse;
	signed long consigne_position;
	signed long pid;
} StructPID;
volatile StructPID PID[2];
unsigned char PID_Actif;

//******************************
// Fonctions
//******************************

void Ports_Init() {					// Configuration du sens des IO
	DDRA = 0;
	DDRB = 0;
	DDRC = 0;
	DDRD = 0;
}
void Mot_Init() {					// Initialisation des moteurs et de la PWM

	// Configuration en sortie des ports
	sbi(MOT_D_IN1_DDR,  MOT_D_IN1_PIN);
	sbi(MOT_D_IN2_DDR,  MOT_D_IN2_PIN);
	sbi(MOT_D_PWM_DDR,  MOT_D_PWM_PIN);
	sbi(MOT_G_IN1_DDR,  MOT_G_IN1_PIN);
	sbi(MOT_G_IN2_DDR,  MOT_G_IN2_PIN);
	sbi(MOT_G_PWM_DDR,  MOT_G_PWM_PIN);

	// Paramtrage du OC1A et OC1B
	TCCR1A = 0b10100011;			// PWM sur 10bit
	TCCR1B = 0b00001001;

	// Immobilisation du robot
	cbi(MOT_D_IN1_PORT,  MOT_D_IN1_PIN);
	cbi(MOT_D_IN2_PORT,  MOT_D_IN2_PIN);
	cbi(MOT_D_PWM_PORT,  MOT_D_PWM_PIN);
	cbi(MOT_G_IN1_PORT,  MOT_G_IN1_PIN);
	cbi(MOT_G_IN2_PORT,  MOT_G_IN2_PIN);
	cbi(MOT_G_PWM_PORT,  MOT_G_PWM_PIN);

	// PWM  0
	OCR1A = 0;
	OCR1B = 0;

}
void PID_Recalc() {					// Calcul des nouvelles commandes

	unsigned char trop = 0;
	unsigned char negatif = 0;

	// Mise  jour des informations
	PID[GAUCHE].vitesse = EncoderState[GAUCHE].position - PID[GAUCHE].position_old;		// La vitesse
	PID[GAUCHE].position_old = EncoderState[GAUCHE].position;
	PID[GAUCHE].erreur_vitesse_old = PID[GAUCHE].erreur_vitesse;						// Sauvegarde de l'ancienne erreur
	PID[GAUCHE].erreur_vitesse = PID[GAUCHE].consigne_vitesse - PID[GAUCHE].vitesse;	// Calcul de la nouvelle erreur
	PID[GAUCHE].integral_vitesse += PID[GAUCHE].erreur_vitesse;							// Calcul du terme intgral

	// Les termes DeltaT et 1/DeltaT sont dj inclus dans Ki et Kd

	// Calcul de la nouvelle commande
	PID[GAUCHE].pid = (signed long)(Kp*(float)PID[GAUCHE].erreur_vitesse + Ki*(float)PID[GAUCHE].integral_vitesse + Kd*((float)PID[GAUCHE].erreur_vitesse-(float)PID[GAUCHE].erreur_vitesse_old));

	// Prise en compte de la nouvelle commande
	if (PID[GAUCHE].pid < 0) {
		negatif = 1;
		cbi(MOT_G_IN1_PORT,  MOT_G_IN1_PIN);
		sbi(MOT_G_IN2_PORT,  MOT_G_IN2_PIN);
		PID[GAUCHE].pid = -PID[GAUCHE].pid;
	} else {
		sbi(MOT_G_IN1_PORT,  MOT_G_IN1_PIN);
		cbi(MOT_G_IN2_PORT,  MOT_G_IN2_PIN);
	}
	if (PID[GAUCHE].pid > 1023) {
		trop = 1;
		PID[GAUCHE].pid = 1023;
	}
	OCR1A = (unsigned int)(PID[GAUCHE].pid);

	// Messages d'extraction de donnes vers le PC
	Usart_putsint((signed int)(PID[GAUCHE].vitesse));
//	Usart_putsint((signed int)(PID[GAUCHE].pid));
	if (negatif)
		Usart_putsint(-(signed int)(OCR1A));
	else
		Usart_putsint((signed int)(OCR1A));
	if (!trop)
		Usart_putc('0');	// Pas de dpassement de tension maximale
	else {
		Usart_putc('1');	// Dpassement de tension maximale : consigne crte
	}

	Usart_putc('\n');
}
void Jauge_Init() {					// Paramtrage de la jauge d'tat

	// Paramtrage du CAN
	ADMUX  = (1<<REFS0)|(JAUGE_CAN_CHANNEL & 0x07);
	ADCSRA = (1<<ADEN)|(1<<ADIE)|(1<<ADIF)|(1<<ADPS2);

	// Paramtrage du Timer 0
	TCCR0 = 0b00000101;				// Prescaler 1:1024 => toutes les 16ms en mode total
	Jauge_overflow = 225;			// On dclenche vite la premire mesure
	sbi(TIMSK, TOIE0);				// Activation
}
void Interrupts_Init() {			// Interruptions diverses

	// Configuration du Timer2 pour les encodeurs
	// Vmax dpend de la frquence pour ne pas rater d'impulsions
	//  10KHz => Vmax = 100cm/s
	//  20KHz => Vmax = 200cm/s
	// f = F_CPU/(Presc*(2^8-init))
	// init = 2^8-F_CPU/(f*Presc)
	TCCR2 = 0b00000010;				// Prescaler = 8 (7.8KHz)
	sbi(TIMSK, TOIE2);				// Activation

	sei();							// Activation des interruptions
}

//******************************
// Interruptions
//******************************

SIGNAL(SIG_ADC) {					// Conversion analogique=>numrique

	// Appel pour mettre  jour la jauge de batterie

	unsigned char adlow, adhigh;
	unsigned int result;
	float rapport;

	// Lecture de la conversion
	adlow  = inp(ADCL);
	adhigh = inp(ADCH);
	result = (adhigh<<8)|adlow;

	// Affichage de l'tat de la batterie
	if (result >= JAUGE_MAXI)
		BarGraph_SetValue(0b1111111111);
	else if (result <= JAUGE_MINI)
		BarGraph_SetValue(0b0000000000);
	else {
		rapport = 10*(result-JAUGE_MINI)/(JAUGE_MAXI-JAUGE_MINI);
		BarGraph_SetValue(0b1111111111>>(10-(char)(rapport)));
	}
}
ISR(TIMER0_OVF_vect) {				// Timer 0, appel toutes les 10ms

	// Recalcul pour 10ms
	TCNT0 = 98;

	count--;

	// Toutes les 50ms :
	if (!count) {
		count = 5;

		// Recalcul de la commande PID
		recalculer = 1;

		if (PID_Actif)
			Duree--;

		// Lancement de la conversion toutes les 0.5s
/*		Jauge_overflow++;
		if (!Jauge_overflow) {
			Jauge_overflow = 245;
			sbi(ADCSRA, ADSC);
		}*/
	}
}
ISR(TIMER2_OVF_vect) {				// Timer 2, appel  15KHz
	TCNT2 = 127;
	Encodeurs_Interrupt();
}

//******************************
// Fonction principale
//******************************

int main() {

	Ports_Init();					// Configuration du sens des IO
	Encodeurs_Init();				// Initialisation des encodeurs
	BarGraph_Init();				// Initialisation du bargraph
	Buffer_Init();					// Initialisation du buffer srie
	Mot_Init();						// Initialisation des moteurs
	Jauge_Init();					// Initialisation de la jauge
	Interrupts_Init();				// Initialisation des interruptions
	Usart_Init();					// Initialisation de la liaison srie

	delay_ms(2000);					// Dlai d'attente aprs mise en route

	// D'abord on ne teste que l'asservissement de type P, de Kp=0  Kp =26
	// Cela permettra de vrifier la mthode de Ziegler Nichols avec les rsultats
	// ultrieurs en PID complet
	Ki = 0;
	Kd = 0;
	for (Kp = 0; Kp < 26; Kp++) {

		Usart_putc('#');
		V_Consigne = 150;
		Duree = 120;
		Usart_putint(Kp);
		Usart_putint(Ki);
		Usart_putint(Kd);
		Usart_putc('\n');
		Usart_putc('\n');

		// Reset des donnes
		EncoderState[GAUCHE].position = 0;
		PID[GAUCHE].position_old = 0;
		PID[GAUCHE].vitesse = 0;
		PID[GAUCHE].erreur_vitesse = 0;
		PID[GAUCHE].erreur_vitesse_old = 0;
		PID[GAUCHE].consigne_vitesse = 0;
		PID[GAUCHE].integral_vitesse = 0;
		PID[GAUCHE].pid = 0;

		// Lancement de la session
		delay_ms(200);
		PID[GAUCHE].consigne_vitesse = V_Consigne;
		PID_Actif = 1;

		while (Duree != 0) {
			if (recalculer) {			// se produit toutes les 50ms
				recalculer = 0;
				PID_Recalc();
				if (recalculer)
					Usart_puts("#erreur timing\n");	// Se produit si la boucle de recalcul dure plus longtemps que l'intervalle d'chantillonage
			}
		}

		PID_Actif = 0;

		// Arrt des moteurs
		OCR1A = 0;

		delay_ms(1000);
	}

	// Maintenant on fait varier simultanment :
	//  - Kp de 2 en 2 de 0  20
	//  - Ki de 5 en 5 de 0  20
	//  - Kd de 0.05 en 0.05 de 0  0.20
	for (Kp = 0; Kp < 21; Kp+=2) {
		for (Ki = 0; Ki < 21; Ki+=5) {
			for (Kd = 0; Kd < 0.21; Kd+=0.05) {

				Usart_putc('#');
				V_Consigne = 150;
				Duree = 120;
				Usart_putint((signed int)(Kp));
				Usart_putint((signed int)(Ki));
				Usart_putint((signed int)(Kd*20));	// Il faudra enlever le *20 logiciellement
				Usart_putc('\n');
				Usart_putc('\n');

				// On inclus DeltaT et 1/DeltaT
				Ki = Ki * 0.05;
				Kd = Kd * 20;

				// Reset des donnes
				EncoderState[GAUCHE].position = 0;
				PID[GAUCHE].position_old = 0;
				PID[GAUCHE].vitesse = 0;
				PID[GAUCHE].erreur_vitesse = 0;
				PID[GAUCHE].erreur_vitesse_old = 0;
				PID[GAUCHE].consigne_vitesse = 0;
				PID[GAUCHE].integral_vitesse = 0;
				PID[GAUCHE].pid = 0;

				// Lancement de la session
				delay_ms(200);
				PID[GAUCHE].consigne_vitesse = V_Consigne;
				PID_Actif = 1;

				while (Duree != 0) {
					if (recalculer) {			// se produit toutes les 50ms
						recalculer = 0;
						PID_Recalc();
						if (recalculer)
							Usart_puts("#erreur timing\n");
					}
				}

				PID_Actif = 0;

				// Arrt des moteurs
				OCR1A = 0;

				// On supprime DeltaT et 1/DeltaT
				Ki = Ki * 20;
				Kd = Kd * 0.05;

				delay_ms(1000);
			}
		}
	}
	while (1)
		delay_ms(100);

	return 0;
}
