/*
 * Robot Pidzy
 *
 * Christophe Le Lann <totofwebcreation@gmail.com>
 * http://www.totofweb.net/robots-projet-55.html
 *
 * Dimensions :
 *  - Rayon roue = 30mm
 *  - Rayon robot = 80mm
 *
 * Conversions :
 *  - 1 pas = pi/32 mm
 *  - 1 pas/interruption = 5pi/8 mm/s
 *  - 90 = 80pi/2 mm = 1280 pas
 *
 * Cette version effectue un test de double asservissement
 * PID ( la fois en vitesse et en position, avec seuillage
 * de la vitesse et de l'acclration) sur deux moteurs,
 * indpendamment l'un de l'autre.
 * Au dbut de la session, les paramtres (Kp, ...) sont
 * demands sur le port srie, puis on lance le test et les
 * donnes sont rcupres par le mme biais pour pouvoir
 * tracer les diffrentes courbes d'volution.
 *
 * Pour simplifier les choses, le PID de vitesse est en ralit
 * un simple P, car les tests ont montrs que a suffisait.
 *
 * Ceci n'est qu'un programme de test et ne suffit pas 
 * raliser un asservissement total d'un robot sur consigne
 * d'une carte matre (il faut modifier un peu le code pour a).
 */

//******************************
// Defines
//******************************

#define F_CPU 16000000UL

//******************************
// Attribution des IO
//******************************

// Encodeurs
#define ENC_G_COUNT_DDR		DDRC
#define ENC_G_COUNT_PORTIN	PINC
#define ENC_G_COUNT_PIN		6
#define ENC_G_SENS_DDR		DDRD
#define ENC_G_SENS_PORTIN	PIND
#define ENC_G_SENS_PIN		2
#define ENC_D_COUNT_DDR		DDRC
#define ENC_D_COUNT_PORTIN	PINC
#define ENC_D_COUNT_PIN		7
#define ENC_D_SENS_DDR		DDRD
#define ENC_D_SENS_PORTIN	PIND
#define ENC_D_SENS_PIN		3

// Moteurs
#define MOT_G_IN1_DDR		DDRC
#define MOT_G_IN1_PORT		PORTC
#define MOT_G_IN1_PIN		5
#define MOT_G_IN2_DDR		DDRC
#define MOT_G_IN2_PORT		PORTC
#define MOT_G_IN2_PIN		4
#define MOT_G_PWM_DDR		DDRD
#define MOT_G_PWM_PORT		PORTD
#define MOT_G_PWM_PIN		4
#define MOT_D_IN1_DDR		DDRC
#define MOT_D_IN1_PORT		PORTC
#define MOT_D_IN1_PIN		3
#define MOT_D_IN2_DDR		DDRC
#define MOT_D_IN2_PORT		PORTC
#define MOT_D_IN2_PIN		2
#define MOT_D_PWM_DDR		DDRD
#define MOT_D_PWM_PORT		PORTD
#define MOT_D_PWM_PIN		5

// BarGraph
#define BARG_RST_DDR		DDRA
#define BARG_RST_PORT		PORTA
#define BARG_RST_PIN		2
#define BARG_CNT_DDR		DDRA
#define BARG_CNT_PORT		PORTA
#define BARG_CNT_PIN		1

// Buffer srie
#define BUFFER_SIZE			128

// Jauge d'tat de la batterie
#define JAUGE_CAN_CHANNEL	0
#define JAUGE_MAXI			149		// Maxi 8.0V, 2^10*U*R/5, avec R=1/11
#define JAUGE_MINI			127		// Mini 6.8V, 2^10*U*R/5, avec R=1/11

//******************************
// Librairies
//******************************

#include <avr/io.h>
#include <avr/interrupt.h>
#include "bitmanip.c"
#include "encodeurs.c"
#include "bargraph.c"
#include "buffer.c"
#include "delais.c"
#include "usart.c"

//******************************
// Variables globales
//******************************

// Jauge d'tat de la batterie
unsigned char Jauge_overflow = 245;	// Compteur divisant par 256 la frquence de mesure
float VKp, VKi, VKd;				// Coefficients de correction pour le PID de vitesse
float PKp;							// Coefficients de correction pour le PID de position (en ralit un P seul)
signed long P_Consigne;				// Consigne  atteindre
unsigned char Amax;					// Acclration maximale
unsigned char Vmax;					// Vitesse maximale
unsigned char recalculer = 0;		// Passe  1 lorsqu'il faut recalculer la correction
unsigned int Duree;					// Duree de session d'enregistrement (en nombre de 50ms)
unsigned char count = 5;			// Variable utilise pour compter les interruptions de 10ms
unsigned char adlow, adhigh;		// Rsultats de la conversion analogique

// PID
typedef struct struct_PID {

	// PID de vitesse
	signed long position_old;
	signed long vitesse;
	signed long erreur_vitesse;
	signed long erreur_vitesse_old;
	signed long consigne_vitesse;
	signed long integral_vitesse;
	signed long pid;

	// Sur-PI de position
	signed long consigne_vitesse_old;
	signed long erreur_position;
	signed long consigne_position;
} StructPID;
volatile StructPID PID[2];
unsigned char PID_Actif = 0;

//******************************
// Fonctions
//******************************

void Ports_Init() {					// Configuration du sens des IO
	DDRA = 0;
	DDRB = 0;
	DDRC = 0;
	DDRD = 0;
}
void Mot_Init() {					// Initialisation des moteurs et de la PWM

	// Configuration en sortie des ports
	sbi(MOT_D_IN1_DDR,  MOT_D_IN1_PIN);
	sbi(MOT_D_IN2_DDR,  MOT_D_IN2_PIN);
	sbi(MOT_D_PWM_DDR,  MOT_D_PWM_PIN);
	sbi(MOT_G_IN1_DDR,  MOT_G_IN1_PIN);
	sbi(MOT_G_IN2_DDR,  MOT_G_IN2_PIN);
	sbi(MOT_G_PWM_DDR,  MOT_G_PWM_PIN);

	// Paramtrage du OC1A et OC1B
	TCCR1A = 0b10100011;			// PWM sur 10bit
	TCCR1B = 0b00001001;

	// Immobilisation du robot
	cbi(MOT_D_IN1_PORT,  MOT_D_IN1_PIN);
	cbi(MOT_D_IN2_PORT,  MOT_D_IN2_PIN);
	cbi(MOT_D_PWM_PORT,  MOT_D_PWM_PIN);
	cbi(MOT_G_IN1_PORT,  MOT_G_IN1_PIN);
	cbi(MOT_G_IN2_PORT,  MOT_G_IN2_PIN);
	cbi(MOT_G_PWM_PORT,  MOT_G_PWM_PIN);

	// PWM  0
	OCR1A = 0;
	OCR1B = 0;

}
void PID_Recalc() {					// Calcul des nouvelles commandes

	unsigned char trop = 0;
	unsigned char negatif = 0;

	// Mise  jour des informations
	PID[GAUCHE].vitesse = EncoderState[GAUCHE].position - PID[GAUCHE].position_old;		// La vitesse
	PID[DROITE].vitesse = EncoderState[DROITE].position - PID[DROITE].position_old;		// La vitesse
	PID[GAUCHE].position_old = EncoderState[GAUCHE].position;
	PID[DROITE].position_old = EncoderState[DROITE].position;

	PID[GAUCHE].erreur_vitesse_old = PID[GAUCHE].erreur_vitesse;						// Sauvegarde de l'ancienne erreur
	PID[DROITE].erreur_vitesse_old = PID[DROITE].erreur_vitesse;						// Sauvegarde de l'ancienne erreur
	PID[GAUCHE].erreur_vitesse = PID[GAUCHE].consigne_vitesse - PID[GAUCHE].vitesse;	// Calcul de la nouvelle erreur
	PID[DROITE].erreur_vitesse = PID[DROITE].consigne_vitesse - PID[DROITE].vitesse;	// Calcul de la nouvelle erreur
	PID[GAUCHE].integral_vitesse += PID[GAUCHE].erreur_vitesse;							// Calcul du terme intgral
	PID[DROITE].integral_vitesse += PID[DROITE].erreur_vitesse;							// Calcul du terme intgral

	// Les termes DeltaT et 1/DeltaT sont dj inclus dans VKi et VKd

	// Calcul de la nouvelle commande
	PID[GAUCHE].pid = (signed long)(VKp*(float)PID[GAUCHE].erreur_vitesse + VKi*(float)PID[GAUCHE].integral_vitesse + VKd*((float)PID[GAUCHE].erreur_vitesse-(float)PID[GAUCHE].erreur_vitesse_old));
	PID[DROITE].pid = (signed long)(VKp*(float)PID[DROITE].erreur_vitesse + VKi*(float)PID[DROITE].integral_vitesse + VKd*((float)PID[DROITE].erreur_vitesse-(float)PID[DROITE].erreur_vitesse_old));

	// Prise en compte de la nouvelle commande
	if (PID[GAUCHE].pid < 0) {
		negatif = 1;
		cbi(MOT_G_IN1_PORT,  MOT_G_IN1_PIN);
		sbi(MOT_G_IN2_PORT,  MOT_G_IN2_PIN);
		PID[GAUCHE].pid = -PID[GAUCHE].pid;
	} else {
		sbi(MOT_G_IN1_PORT,  MOT_G_IN1_PIN);
		cbi(MOT_G_IN2_PORT,  MOT_G_IN2_PIN);
	}
	if (PID[DROITE].pid < 0) {
//		negatif = 1;
		cbi(MOT_D_IN1_PORT,  MOT_D_IN1_PIN);
		sbi(MOT_D_IN2_PORT,  MOT_D_IN2_PIN);
		PID[DROITE].pid = -PID[DROITE].pid;
	} else {
		sbi(MOT_D_IN1_PORT,  MOT_D_IN1_PIN);
		cbi(MOT_D_IN2_PORT,  MOT_D_IN2_PIN);
	}
	if (PID[GAUCHE].pid > 1023) {
		trop = 1;
		PID[GAUCHE].pid = 1023;
	}
	if (PID[DROITE].pid > 1023) {
//		trop = 1;
		PID[DROITE].pid = 1023;
	}
	OCR1A = (unsigned int)(PID[GAUCHE].pid);
	OCR1B = (unsigned int)(PID[DROITE].pid);

	// Messages d'extraction de donnes vers le PC
	Usart_putsint((signed int)(PID[GAUCHE].vitesse));
//	Usart_putsint((signed int)(PID[GAUCHE].pid));
	if (negatif)
		Usart_putsint(-(signed int)(OCR1A));
	else
		Usart_putsint((signed int)(OCR1A));
	if (!trop)
		Usart_putc('0');	// Pas de dpassement de tension maximale
	else {
		Usart_putc('1');	// Dpassement de tension maximale : consigne crte
	}

	Usart_putc('\n');
}
void Jauge_Init() {					// Paramtrage de la jauge d'tat

	// Paramtrage du CAN
	ADMUX  = (1<<REFS0)|(JAUGE_CAN_CHANNEL & 0x07);
	ADCSRA = (1<<ADEN)|(1<<ADIE)|(1<<ADIF)|(1<<ADPS2);

	// Paramtrage du Timer 0
	TCCR0 = 0b00000101;				// Prescaler 1:1024 => toutes les 16ms en mode total
	Jauge_overflow = 225;			// On dclenche vite la premire mesure
	sbi(TIMSK, TOIE0);				// Activation
}
void Jauge_Start() {				// Lancer la conversion AN
	sbi(ADCSRA, ADSC);
}
void Jauge_Aff() {					// Affichage de l'tat de la batterie

	float rapport;
	unsigned int result;

	if (PID_Actif)					// Scurit pour ne pas fausser les rsultats
		return;

	result = (adhigh<<8)|(adlow);
	if (result >= JAUGE_MAXI)
		BarGraph_SetValue(0b1111111111);
	else if (result <= JAUGE_MINI)
		BarGraph_SetValue(0b0000000000);
	else {
		rapport = 10*(result-JAUGE_MINI)/(JAUGE_MAXI-JAUGE_MINI);
		BarGraph_SetValue(0b1111111111>>(10-(char)(rapport)));
	}
}
void Interrupts_Init() {			// Interruptions diverses

	// Configuration du Timer2 pour les encodeurs
	// Vmax dpend de la frquence pour ne pas rater d'impulsions
	//  10KHz => Vmax = 100cm/s
	//  20KHz => Vmax = 200cm/s
	// f = F_CPU/(Presc*(2^8-init))
	// init = 2^8-F_CPU/(f*Presc)
	TCCR2 = 0b00000010;				// Prescaler = 8 (7.8KHz)
	sbi(TIMSK, TOIE2);				// Activation

	sei();							// Activation des interruptions
}

//******************************
// Interruptions
//******************************

SIGNAL(SIG_ADC) {					// Conversion analogique=>numrique

	// Lecture de la conversion
	adlow  = inp(ADCL);
	adhigh = inp(ADCH);

	if (!PID_Actif)
		Jauge_Aff();
}
ISR(TIMER0_OVF_vect) {				// Timer 0, appel toutes les 10ms

	// Recalcul pour 10ms
	TCNT0 = 98;

	count--;

	// Toutes les 50ms :
	if (!count) {
		count = 5;

		// Recalcul de la commande PID
		recalculer = 1;

		if (PID_Actif)
			Duree--;

		// Lancement de la conversion toutes les 0.5s
		Jauge_overflow++;
		if (!Jauge_overflow) {
			Jauge_overflow = 245;
			Jauge_Start();
		}
	}
}
ISR(TIMER2_OVF_vect) {				// Timer 2, appel  15KHz
	TCNT2 = 127;
	Encodeurs_Interrupt();
}

//******************************
// Fonction principale
//******************************

int main() {

	Ports_Init();					// Configuration du sens des IO
	Encodeurs_Init();				// Initialisation des encodeurs
	BarGraph_Init();				// Initialisation du bargraph
	Buffer_Init();					// Initialisation du buffer srie
	Mot_Init();						// Initialisation des moteurs
	Jauge_Init();					// Initialisation de la jauge
	Interrupts_Init();				// Initialisation des interruptions
	Usart_Init();					// Initialisation de la liaison srie

	delay_ms(2000);					// Dlai d'attente aprs mise en route

	// Message signalant tout restart
	Usart_puts("\n## CartoPID - Demarrage\n");

	while(1) {

		Usart_puts("# Double asservissement vitesse-position\n");

		// Menu de demande des paramtres de la session
		VKp = Usart_GetParam("\n# V Kp(2)=");   // Valeur recommande = 2
		VKi = Usart_GetParam("\n# V Ki(3)=");   // Valeur recommande = 3
		VKd = Usart_GetParam("\n# V Kd(3)=");   // Valeur recommande = 3
		PKp = Usart_GetParam("\n# P Kp(0.2)="); // Valeur recommande = 0.2
		Amax = (unsigned char)Usart_GetParam("\n# Amax(120)="); // Valeur recommande = 120 (soit ~24cm/s)
		Vmax = (unsigned char)Usart_GetParam("\n# Vmax(150)="); // Valeur recommande = 150 (soit ~30cm/s)
		PID[GAUCHE].consigne_position = (signed long)Usart_GetParam("\n# P Consigne=");
		if ((unsigned char)Usart_GetParam("\n# {0:Droit,1:RotD}=") == 1)
			PID[DROITE].consigne_position = -PID[GAUCHE].consigne_position;
		else
			PID[DROITE].consigne_position = PID[GAUCHE].consigne_position;
		Duree = (unsigned int)Usart_GetParam("\n# Duree acquisition (s)=");
		Usart_puts("\n\n");

		// Adaptation de certains coefficients pour DeltaT = 50ms
		Amax *= 0.05;		// Passage de pas/s en pas/(0.05s)
		Duree *= 20;		// On transpose en nombre de 50ms

		// Reset des donnes
		EncoderState[GAUCHE].position = 0;
		EncoderState[DROITE].position = 0;
		PID[GAUCHE].position_old = 0;
		PID[DROITE].position_old = 0;
		PID[GAUCHE].vitesse = 0;
		PID[DROITE].vitesse = 0;
		PID[GAUCHE].erreur_vitesse = 0;
		PID[DROITE].erreur_vitesse = 0;
		PID[GAUCHE].erreur_vitesse_old = 0;
		PID[DROITE].erreur_vitesse_old = 0;
		PID[GAUCHE].consigne_vitesse = 0;
		PID[DROITE].consigne_vitesse = 0;
		PID[GAUCHE].consigne_vitesse_old = 0;
		PID[DROITE].consigne_vitesse_old = 0;
		PID[GAUCHE].integral_vitesse = 0;
		PID[DROITE].integral_vitesse = 0;
		PID[GAUCHE].pid = 0;
		PID[DROITE].pid = 0;
		PID[GAUCHE].erreur_position = 0;
		PID[DROITE].erreur_position = 0;

		// Lancement de la session
		delay_ms(200);
		PID_Actif = 1;

		while (Duree != 0) {
			if (recalculer) {			// se produit toutes les 50ms
				recalculer = 0;

				PID[GAUCHE].erreur_position = PID[GAUCHE].consigne_position - EncoderState[GAUCHE].position;	// Calcul de la nouvelle erreur de position
				PID[DROITE].erreur_position = PID[DROITE].consigne_position - EncoderState[DROITE].position;	// Calcul de la nouvelle erreur de position
				PID[GAUCHE].consigne_vitesse_old = PID[GAUCHE].consigne_vitesse;
				PID[DROITE].consigne_vitesse_old = PID[DROITE].consigne_vitesse;
				PID[GAUCHE].consigne_vitesse = (signed long)(PKp*(float)PID[GAUCHE].erreur_position);
				PID[DROITE].consigne_vitesse = (signed long)(PKp*(float)PID[DROITE].erreur_position);

				// Ecrtage de la vitesse
				if (PID[GAUCHE].consigne_vitesse >= (signed long)Vmax) {
					PID[GAUCHE].consigne_vitesse = (signed long)Vmax;
				} else if (PID[GAUCHE].consigne_vitesse <= -(signed long)Vmax) {
					PID[GAUCHE].consigne_vitesse = -(signed long)Vmax;
				}
				if (PID[DROITE].consigne_vitesse >= (signed long)Vmax) {
					PID[DROITE].consigne_vitesse = (signed long)Vmax;
				} else if (PID[DROITE].consigne_vitesse <= -(signed long)Vmax) {
					PID[DROITE].consigne_vitesse = -(signed long)Vmax;
				}

				// Ecrtage de l'acclration
				if ((PID[GAUCHE].consigne_vitesse - PID[GAUCHE].consigne_vitesse_old) >= (signed long)Amax) {
					PID[GAUCHE].consigne_vitesse = PID[GAUCHE].consigne_vitesse_old + Amax;
				} else if ((PID[GAUCHE].consigne_vitesse - PID[GAUCHE].consigne_vitesse_old) <= -(signed long)Amax) {
					PID[GAUCHE].consigne_vitesse = PID[GAUCHE].consigne_vitesse_old - Amax;
				}
				if ((PID[DROITE].consigne_vitesse - PID[DROITE].consigne_vitesse_old) >= (signed long)Amax) {
					PID[DROITE].consigne_vitesse = PID[DROITE].consigne_vitesse_old + Amax;
				} else if ((PID[DROITE].consigne_vitesse - PID[DROITE].consigne_vitesse_old) <= -(signed long)Amax) {
					PID[DROITE].consigne_vitesse = PID[DROITE].consigne_vitesse_old - Amax;
				}

				Usart_putsint((signed int)EncoderState[GAUCHE].position);
				Usart_putsint((signed int)PID[GAUCHE].consigne_vitesse);

				PID_Recalc();

				if (recalculer)
					Usart_puts("#erreur timing\n");
			}
		}

		PID_Actif = 0;

		// Arrt des moteurs
		OCR1A = 0;
		OCR1B = 0;

		// Doit-on recommencer ?
		Usart_GetParam("\n# Nouvelle session ?\n");
	}

	return 0;
}
